/*!
 * © 2021 David Vines
 *
 * Creative Commons License
 *
 * This work is licensed under a Creative Commons Attribution-NonCommercial-ShareAlike 4.0 International License.
 */

/* globals ROT, Result */
/* exported Opening, Contination, Bout */
class Opening {
	constructor(name) {
		this._name = name;
	}
	toString() {
		return this._name;
	}
	getOdds(theirMove,us,them) {
		throw new Error("getOdds not provided for "+this.toString()+".getOdds("+theirMove.toString()+","+us.toString()+","+them.toString()+")");
	}
	static _getOdds(defence,ourstat,theirstat) {
		const offense = 40;
		return (ourstat+offense)/(ourstat+offense+theirstat+defence)*100;
	}
	static neitherSucceeded(result) {
		result.addText("Both rikishi crash together from the tachi-ai and lock up in the middle of the dohyo with neither having the advantage.");
	}
}
Opening.Charge = new Opening("Charge");
Opening.Charge.getOdds = function(theirMove,us,them) {
	switch(true) {
		case theirMove === Opening.Charge:
			return Opening._getOdds(85,us.getSpeed(),them.getWeight());
		case theirMove === Opening.Attack:
			return Opening._getOdds(-20,us.getSpeed(),them.getWeight());
		case theirMove === Opening.StepAside:
			return Opening._getOdds(460,us.getSpeed(),them.getWeight());
		case theirMove === Opening.Slap:
			return Opening._getOdds(10,us.getSpeed(),them.getWeight());
		default:
			throw new Error("Unknown countermove "+theirMove.toString()+" when determing odds for Opening.Charge");
	}
};
Opening.Charge.bothSameProhibited = function(result,us,them) {
	result.addText("Both "+us.toString()+" and "+them.toString()+" charge.");
	return false;
};
Opening.Charge.addBothTriedMoveAndSucceededDescription = function(result) {
	result.addText("They crash heavily together in the middle of the dohyo.");
};
Opening.Charge.addSuccessfulDescription = function(result,us,them,_theirSuccessfulMove) {
	result.addText(us.toString()+" charges "+them.toString()+" but "+them.toString()+" absorbs the charge.");
};
Opening.Charge.addWinDescription = function(result,us,them,_theirMove) {
	result.addText(us.toString()+" charges "+them.toString()+" pushing "+them.toString()+" all the way to the edge of the dohyo and out. "+us.toString()+" wins by Oshi-dashi.");
};
Opening.Attack = new Opening("Attack");
Opening.Attack.getOdds = function(theirMove,us,them) {
	switch(true) {
		case theirMove === Opening.Charge:
			return Opening._getOdds(460,us.getStrength(),them.getStrength());
		case theirMove === Opening.Attack:
			return Opening._getOdds(85,us.getStrength(),them.getStrength());
		case theirMove === Opening.StepAside:
			return Opening._getOdds(-20,us.getStrength(),them.getStrength());
		case theirMove === Opening.Slap:
			return Opening._getOdds(60,us.getStrength(),them.getStrength());
		default:
			throw new Error("Unknown countermove "+theirMove.toString()+" when determing odds for Opening.Attack");
	}
};
Opening.Attack.bothSameProhibited = function(result,us,them) {
	result.addText("Both "+us.toString()+" and "+them.toString()+" go for the belt.");
	return false;
};
Opening.Attack.addBothTriedMoveAndSucceededDescription = function(result) {
	result.addText("They each get a grip on each other's belt.");
};
Opening.Attack.addSuccessfulDescription = function(result,us,them,_theirSuccessfulMove) {
	result.addText(us.toString()+" grabs the belt of "+them.toString()+".");
};
Opening.Attack.addWinDescription = function(result,us,them,_theirMove) {
	result.addText(us.toString()+" grabs the belt of "+them.toString()+" and pushes "+them.toString()+" all the way to the edge of the dohyo and out of it. "+us.toString()+" wins by Yori-kiri.");
};
Opening.StepAside = new Opening("Step Aside");
Opening.StepAside.getOdds = function(theirMove,us,them) {
	switch(true) {
		case theirMove === Opening.Charge:
			return Opening._getOdds(-20,us.getAgility(),them.getSpeed());
		case theirMove === Opening.Attack:
			return Opening._getOdds(460,us.getAgility(),them.getSpeed());
		case theirMove === Opening.StepAside:
			return Opening._getOdds(85,us.getAgility(),them.getSpeed());
		case theirMove === Opening.Slap:
			return Opening._getOdds(160,us.getAgility(),them.getSpeed());
		default:
			throw new Error("Unknown countermove "+theirMove.toString()+" when determing odds for Opening.StepAside");
	}
};
Opening.StepAside.bothSameProhibited = function(_result,_us,_them) {
	return true;
};
Opening.StepAside.addBothTriedMoveAndSucceededDescription = function(_result) {
	throw new Error("Both rikishi succeeded in a step aside move - which the code should have prevented");
};
Opening.StepAside.addSuccessfulDescription = function(result,us,them,theirSuccessfulMove) {
	if (theirSuccessfulMove === Opening.Charge) {
		result.addText(us.toString()+" steps to the side of "+them.toString()+".");
	} else {
		result.addText(us.toString()+" steps to the side and behind "+them.toString()+".");
	}
};
Opening.StepAside.addWinDescription = function(result,us,them,_theirMove) {
	result.addText(us.toString()+" steps to the side and behind "+them.toString()+" pushing "+them.toString()+" out of the dohyo. "+us.toString()+" wins by Okuri-dashi.");
};
Opening.Slap = new Opening("Slap");
Opening.Slap.getOdds = function(theirMove,us,them) {
	switch(true) {
		case theirMove === Opening.Charge:
			return Opening._getOdds(160,us.getStrength(),them.getWeight());
		case theirMove === Opening.Attack:
			return Opening._getOdds(60,us.getStrength(),them.getWeight());
		case theirMove === Opening.StepAside:
			return Opening._getOdds(10,us.getStrength(),them.getWeight());
		case theirMove === Opening.Slap:
			return Opening._getOdds(85,us.getStrength(),them.getWeight());
		default:
			throw new Error("Unknown countermove "+theirMove.toString()+" when determing odds for Opening.Slap");
	}
};
Opening.Slap.bothSameProhibited = function(result,us,them) {
	result.addText("Both "+us.toString()+" and "+them.toString()+" come out slapping.");
	return false;
};
Opening.Slap.addBothTriedMoveAndSucceededDescription = function(_result) {
};
Opening.Slap.addSuccessfulDescription = function(result,us,them,_theirSuccessfulMove) {
	result.addText(us.toString()+" slaps "+them.toString()+".");
};
Opening.Slap.addWinDescription = function(result,us,them,theirMove) {
	if (theirMove === Opening.Charge) {
		result.addText(us.toString()+" slaps down "+them.toString()+" as they charge, winning by Hataki-komi.");
	} else {
		result.addText(us.toString()+" slaps down "+them.toString()+" winning by Tsuki-dashi.");
	}
};
Opening.eastmoves = [Opening.Charge,Opening.Attack,Opening.StepAside,Opening.Slap];
Opening.westmoves = Opening.eastmoves;

class Continuation {
	constructor(name) {
		this._name = name;
	}
	toString() {
		return this._name;
	}
	getOdds(theirMove,us,them) {
		throw new Error("getOdds not provided for "+this.toString()+".getOdds("+theirMove.toString()+","+us.toString()+","+them.toString()+")");
	}
	bothSameProhibited(_result,_us,_them) {
		return false;
	}
	static _getOdds(defence,ourstat,theirstat) {
		const offense = 40;
		return (ourstat+offense)/(ourstat+offense+theirstat+defence)*100;
	}
	static neitherSucceeded(result) {
		result.addText("Both rikishi remain locked up in the middle of the dohyo with neither having the advantage.");
	}
}
Continuation.Lever = new Continuation("Throw");
Continuation.Lever.getOdds = function(theirMove,us,them) {
	switch(true) {
		case theirMove === Continuation.Lever:
			return Continuation._getOdds(85,us.getAgility(),them.getStrength());
		case theirMove === Continuation.Lift:
			return Continuation._getOdds(460,us.getAgility(),them.getStrength());
		case theirMove === Continuation.Push:
			return Continuation._getOdds(-20,us.getAgility(),them.getStrength());
		case theirMove === Continuation.Pull:
			return Continuation._getOdds(10,us.getAgility(),them.getStrength());
		case theirMove === Continuation.Trip:
			return Continuation._getOdds(10,us.getAgility(),them.getStrength());
		default:
			throw new Error("Unknown countermove "+theirMove.toString()+" when determing odds for Continuation.Lever");
	}
};
Continuation.Lever.addBothTriedMoveAndSucceededDescription = function(result) {
	result.addText("Both rikishi attempt to throw their opponent, but neither gains the advantage.");
};
Continuation.Lever.addSuccessfulDescription = function(result,us,them,_theirSuccessfulMove) {
	result.addText(us.toString()+" tries to throw "+them.toString()+".");
};
Continuation.Lever.addWinDescription = function(result,us,them,_theirMove) {
	result.addText(us.toString()+" throws "+them.toString()+". "+us.toString()+" wins by Uwate-nage.");
};
Continuation.Lift = new Continuation("Lift");
Continuation.Lift.getOdds = function(theirMove,us,them) {
	switch(true) {
		case theirMove === Continuation.Lever:
			return Continuation._getOdds(-20,us.getStrength(),them.getWeight());
		case theirMove === Continuation.Lift:
			return Continuation._getOdds(85,us.getStrength(),them.getWeight());
		case theirMove === Continuation.Push:
			return Continuation._getOdds(460,us.getStrength(),them.getWeight());
		case theirMove === Continuation.Pull:
			return Continuation._getOdds(60,us.getStrength(),them.getWeight());
		case theirMove === Continuation.Trip:
			return Continuation._getOdds(-20,us.getStrength(),them.getWeight());
		default:
			throw new Error("Unknown countermove "+theirMove.toString()+" when determing odds for Continuation.Lift");
	}
};
Continuation.Lift.addBothTriedMoveAndSucceededDescription = function(result) {
	result.addText("Both rikishi attempt to lift their opponent, but neither gains the advantage.");
};
Continuation.Lift.addSuccessfulDescription = function(result,us,them,_theirSuccessfulMove) {
	result.addText(us.toString()+" attempts to lift "+them.toString()+".");
};
Continuation.Lift.addWinDescription = function(result,us,them,_theirMove) {
	result.addText(them.toString()+" is lifted out of the dohyo and "+us.toString()+" wins by Yori-taoshi.");
};
Continuation.Push = new Continuation("Push");
Continuation.Push.getOdds = function(theirMove,us,them) {
	switch(true) {
		case theirMove === Continuation.Lever:
			return Continuation._getOdds(460,us.getWeight(),them.getWeight());
		case theirMove === Continuation.Lift:
			return Continuation._getOdds(-20,us.getWeight(),them.getWeight());
		case theirMove === Continuation.Push:
			return Continuation._getOdds(85,us.getWeight(),them.getWeight());
		case theirMove === Continuation.Pull:
			return Continuation._getOdds(160,us.getWeight(),them.getWeight());
		case theirMove === Continuation.Trip:
			return Continuation._getOdds(160,us.getWeight(),them.getWeight());
		default:
			throw new Error("Unknown countermove "+theirMove.toString()+" when determing odds for Continuation.Push");
	}
};
Continuation.Push.addBothTriedMoveAndSucceededDescription = function(result) {
	result.addText("Both rikishi attempt to push their opponent out of the dohyo, but neither gains the advantage.");
};
Continuation.Push.addSuccessfulDescription = function(result,us,them,_theirSuccessfulMove) {
	result.addText(us.toString()+" pushs "+them.toString()+", but "+them.toString()+" remains in the dohyo.");
};
Continuation.Push.addWinDescription = function(result,us,them,_theirMove) {
	result.addText(them.toString()+" is pushed out of the dohyo and "+us.toString()+" wins by Oshi-dashi.");
};
Continuation.Pull = new Continuation("Pull");
Continuation.Pull.getOdds = function(theirMove,us,them) {
	switch(true) {
		case theirMove === Continuation.Lever:
			return Continuation._getOdds(160,us.getStrength(),them.getAgility());
		case theirMove === Continuation.Lift:
			return Continuation._getOdds(60,us.getStrength(),them.getAgility());
		case theirMove === Continuation.Push:
			return Continuation._getOdds(10,us.getStrength(),them.getAgility());
		case theirMove === Continuation.Pull:
			return Continuation._getOdds(85,us.getStrength(),them.getAgility());
		case theirMove === Continuation.Trip:
			return Continuation._getOdds(85,us.getStrength(),them.getAgility());
		default:
			throw new Error("Unknown countermove "+theirMove.toString()+" when determing odds for Continuation.Pull");
	}
};
Continuation.Pull.addBothTriedMoveAndSucceededDescription = function(result) {
	result.addText("Both rikishi attempt to pull their opponent to the ground, but neither gains the advantage.");
};
Continuation.Pull.addSuccessfulDescription = function(result,us,them,_theirSuccessfulMove) {
	result.addText(us.toString()+" pulls on "+them.toString()+".");
};
Continuation.Pull.addWinDescription = function(result,us,them,_theirMove) {
	result.addText(them.toString()+" is pulled to the clay and "+us.toString()+" wins by Hiki-otoshi.");
};
Continuation.Trip = new Continuation("Trip");
Continuation.Trip.getOdds = function(theirMove,us,them) {
	switch(true) {
		case theirMove === Continuation.Lever:
			return Continuation._getOdds(460,us.getSpeed(),them.getAgility());
		case theirMove === Continuation.Lift:
			return Continuation._getOdds(-20,us.getSpeed(),them.getAgility());
		case theirMove === Continuation.Push:
			return Continuation._getOdds(10,us.getSpeed(),them.getAgility());
		case theirMove === Continuation.Pull:
			return Continuation._getOdds(460,us.getSpeed(),them.getAgility());
		case theirMove === Continuation.Trip:
			return Continuation._getOdds(85,us.getSpeed(),them.getAgility());
		default:
			throw new Error("Unknown countermove "+theirMove.toString()+" when determing odds for Continuation.Trip");
	}
};
Continuation.Trip.addBothTriedMoveAndSucceededDescription = function(result) {
	result.addText("Both rikishi attempt to trip their opponent, but neither gains the advantage.");
};
Continuation.Trip.addSuccessfulDescription = function(result,us,them,_theirSuccessfulMove) {
	result.addText(us.toString()+" tries to trip "+them.toString()+".");
};
Continuation.Trip.addWinDescription = function(result,us,them,_theirMove) {
	result.addText(us.toString()+" trips "+them.toString()+" and "+us.toString()+" wins by Soto-gake.");
};
Continuation.eastmoves = [Continuation.Lever,Continuation.Lift,Continuation.Push,Continuation.Pull,Continuation.Trip];
Continuation.westmoves = Continuation.eastmoves;

class Bout {
	static resolve(east,west) {
		let time = 0;
		const result = new Result();
		let state = Bout._resolveMoves(Opening,result,east,west,time++);
		while(result.getWinner() === null) {
			result.newline();
			state = Bout._resolveMoves(state,result,east,west,time++);
		}
		return result;
	}
	static _resolveMoves(state,result,east,west,time) {
		let eastmove, westmove;
		do {
			eastmove = east.getMove(west,state.eastmoves,time,result);
			westmove = west.getMove(east,state.westmoves,time,result);
		}
		while(eastmove === westmove && eastmove.bothSameProhibited(result,east,west));
		//result.newline();
		//result.addText(`${east} ${eastmove} ${Math.floor(eastmove.getOdds(westmove,east,west))}%, ${west} ${westmove} ${Math.floor(westmove.getOdds(eastmove,west,east))}%`);
		//result.addText(`${east} saves ${Math.floor(east.getSaveChance(time))}%, ${west} saves ${Math.floor(west.getSaveChance(time))}%`);
		//result.newline();
		const eastmovesuccess = (ROT.RNG.getPercentage() <= eastmove.getOdds(westmove,east,west));
		const westmovesuccess = (ROT.RNG.getPercentage() <= westmove.getOdds(eastmove,west,east));
		const eastsaves = east.saves(time);
		const westsaves = west.saves(time);
		const eastposswin = eastmovesuccess && !westsaves;
		const westposswin = westmovesuccess && !eastsaves;
		if (eastposswin && westposswin) {
			// Both rikishi succeeded and both failed to save - Need to determine who was the fastest
			const eastspeed = ROT.RNG.getPercentage()+east.getBalance()/2+ROT.RNG.getUniform()/5;
			const westspeed = ROT.RNG.getPercentage()+west.getBalance()/2+ROT.RNG.getUniform()/5;
			if (eastspeed > westspeed) {
				eastmove.addWinDescription(result,east,west,westmove);
				result.setWinner(east);
				result.setLoser(west);
			} else {
				westmove.addWinDescription(result,west,east,eastmove);
				result.setWinner(west);
				result.setLoser(east);
			}
		} else if (eastposswin) {
			eastmove.addWinDescription(result,east,west,westmove);
			result.setWinner(east);
			result.setLoser(west);
		} else if (westposswin) {
			westmove.addWinDescription(result,west,east,eastmove);
			result.setWinner(west);
			result.setLoser(east);
		} else if (eastmovesuccess && westmovesuccess) {
			// Both rikishi succeeded in their moves, but also both saved
			if (eastmove === westmove) {
				// Both made the same move
				eastmove.addBothTriedMoveAndSucceededDescription(result);
			} else {
				eastmove.addSuccessfulDescription(result,east,west,westmove);
				westmove.addSuccessfulDescription(result,west,east,eastmove);
			}
		} else if (eastmovesuccess) {
				eastmove.addSuccessfulDescription(result,east,west,null);
		} else 	if (westmovesuccess) {
				westmove.addSuccessfulDescription(result,west,east,null);
		} else {
			// Neither rikishi succeeded
			state.neitherSucceeded(result);
		}

		return Continuation;
	}
}